//*****************************************************************************
//
// tm4c129_onewire_enum.c - Example demonstrating how to configure the 1-Wire
//           controller on TM4C129XNCZAD and enumerate 1-wire devices.
//
// Copyright (c) 2014 Texas Instruments Incorporated.  All rights reserved.
// Software License Agreement
//
//   Redistribution and use in source and binary forms, with or without
//   modification, are permitted provided that the following conditions
//   are met:
//
//   Redistributions of source code must retain the above copyright
//   notice, this list of conditions and the following disclaimer.
//
//   Redistributions in binary form must reproduce the above copyright
//   notice, this list of conditions and the following disclaimer in the
//   documentation and/or other materials provided with the
//   distribution.
//
//   Neither the name of Texas Instruments Incorporated nor the names of
//   its contributors may be used to endorse or promote products derived
//   from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
//*****************************************************************************

#include <stdint.h>
#include <stdbool.h>
#include "inc/hw_types.h"
#include "inc/hw_gpio.h"
#include "inc/hw_ints.h"
#include "inc/hw_memmap.h"
#include "inc/hw_sysctl.h"
#include "inc/hw_onewire.h"
#include "driverlib/interrupt.h"
#include "driverlib/sysctl.h"
#include "driverlib/onewire.h"
#include "driverlib/gpio.h"
#include "driverlib/pin_map.h"
#include "driverlib/uart.h"
#include "utils/uartstdio.h"
#include "drivers/pinout.h"

//*****************************************************************************
//
// Define for Maximum number of devices that can be searched
//
//*****************************************************************************
#define MAX_DEVICES 30

//*****************************************************************************
//
// Define for Search ROM COMMAND
//
//*****************************************************************************
#define SEARCH_ROM_COMMAND 0xF0

//*****************************************************************************
//
// Define for Binary Search Path
//
//*****************************************************************************
#define TAKE_PATH1 1
#define TAKE_PATH0 0

//*****************************************************************************
//
// Define for CRC 9 Polynomial
// Polynomial used is : x^8+x^5+x^4+1
//
//*****************************************************************************
#define CRC_POLY9 0b100110001

//*****************************************************************************
//
// Poynomial Degree
//
//*****************************************************************************
#define POLY_DEGREE 8

//*****************************************************************************
//
// Global Variable Declaration for the address of One Wire devices on the bus
//
//*****************************************************************************
volatile uint64_t g_pui64SlaveAddr[MAX_DEVICES];

//*****************************************************************************
//
//! Performs CRC check.
//!
//! \param ui64SlaveAddr gives the data to check CRC upon.
//!
//! \return 1 if CRC is correct. 0 if CRC is incorrect.
//
//*****************************************************************************
uint32_t
OneWireCrcCheck(uint64_t ui64SlaveAddr)
{
    int32_t i32StartBitPosition ;
    int32_t i32PolyDegreeNumber ;

    for(i32StartBitPosition = 0 ; i32StartBitPosition <= 55 ;
            i32StartBitPosition++)
    {
        //
        // If there is a '0' in i32StartBitPosition then there are
        // no changes to be made in ull_addr_test.
        //
        if((( ui64SlaveAddr >> i32StartBitPosition ) & 1 ) != 1)
        {
            continue; // Continue from i32StartBitPosition.
        }

        for(i32PolyDegreeNumber = 0 ; i32PolyDegreeNumber <= POLY_DEGREE
                                                    ; i32PolyDegreeNumber++)
        {
            //
            // Implementing XOR of bits in polynomial
            // and corresponding bits in ull_addr_test.
            //
            if (((CRC_POLY9 >> (POLY_DEGREE - i32PolyDegreeNumber)) & 1) ==
                ((ui64SlaveAddr >> ((uint64_t)i32StartBitPosition +
                        (uint64_t)i32PolyDegreeNumber)) & 1))
            {
                ui64SlaveAddr &= (0xffffffffffffffff -
                    ((uint64_t)1 <<
                            (i32StartBitPosition + i32PolyDegreeNumber)));
            }
            else
            {

                ui64SlaveAddr |= ((uint64_t)1 <<
                        (i32StartBitPosition + i32PolyDegreeNumber));
            }
        }
    }

    //
    // Check if CRC obtained by the algorithm
    // and CRC value given by device are same.
    //
    if (ui64SlaveAddr != 0)
    {
        //
        // Return 0 ==> Wrong CRC
        //
        return 0 ;
    }

    //
    // Successful CRC check returns 1
    //
    return 1;
}

//*****************************************************************************
//
//  Searches for devices on the bus using binary tree algorithm
//
//      This function updates g_pui64SlaveAddr array with the slave
//      addresses and number of slaves
//
//      \return number of slaves found. Return 0 ==> no devices found
//
//
//*****************************************************************************
uint32_t
OneWireEnum(void)
{
    //
    // Contains the value of two read bits that are to be read in
    // each iteration while obtaining a ROM number.
    //
    uint32_t ui32ReadBits;

    //
    // When there is a conflict and ui32BitNumber less than
    // i32ConflictBitNumber, then ui32WriteBit (path of last ROM number)
    // contains the path that has to be taken.
    //
    uint32_t ui32WriteBit;

    //
    // Number of the bit under cosideration in the perticular iteration.
    //
    uint32_t ui32BitNumber;

    //
    // Temporary variable to store slave address while finding it.
    //
    uint64_t ui64SlaveAddr;

    //
    // If there is a conflict, the path to be taken is decided based on
    // value of i32ConflictBitNumber.
    // The value of i32LastConflictZeroBitNumber after finding the
    // previous ROM number is stored in i32ConflictBitNumber.
    // i32ConflictBitNumber is initiated to -1, because while finding
    // first ROM number on the bus there is no previous ROM number to
    // refer to, which is a case if there is a conflict at
    // ui32BitNumber less than i32ConflictBitNumber.
    //
    int32_t i32ConflictBitNumber = -1 ;

    //
    // At the start, there are no devices found
    //
    int32_t i32NumSlaves = 0;

    //
    // While finding a ROM number, if there is a conflict
    // and the path taken is zero then the value of
    // the  ui32BitNumber is stored in i32LastConflictZeroBitNumber.
    //
    int32_t i32LastConflictZeroBitNumber = 0;

    //
    // Issue Reset
    //
    OneWireBusReset(ONEWIRE0_BASE);

    //
    // Wait for the completion of reset
    //
    while((OneWireIntStatus(ONEWIRE0_BASE, false)
                      & ONEWIRE_INT_RESET_DONE) != 0);

    //
    // Clear ONEWIRE_INT_RESET_DONE interrupt
    //
    OneWireIntClear(ONEWIRE0_BASE, ONEWIRE_INT_RESET_DONE);

    if((OneWireIntStatus(ONEWIRE0_BASE, false)
                     & ONEWIRE_INT_NO_SLAVE) != 0)
    {
        //
        // Return 0 ==> no devices found
        //
        return 0;
    }

    //
    // Search for devices till no more undetected devices are present
    // or till the maximum no of devices are reached
    //
    while((i32LastConflictZeroBitNumber != -1)
                 && (i32NumSlaves < MAX_DEVICES))
    {
        //
        // i32LastConflictZeroBitNumber is set to -1 at the start
        // searching for each ROM number.
        // No update to i32LastConflictZeroBitNumber while finding
        // a ROM number means that there are no more undetected devices,
        // which is used as a
        // condition for while loop.
        //
        i32LastConflictZeroBitNumber = -1;

        //
        // Reset before start searching for new device
        //
        OneWireBusReset(ONEWIRE0_BASE);

        //
        // Wait for the completion of reset
        //
        while((OneWireIntStatus(ONEWIRE0_BASE, false)
                         & ONEWIRE_INT_RESET_DONE) == 0);

        //
        // Clear ONEWIRE_INT_RESET_DONE interrupt
        //
        OneWireIntClear(ONEWIRE0_BASE, ONEWIRE_INT_RESET_DONE);

        //
        // Write Search ROM Command
        //
        OneWireTransaction(ONEWIRE0_BASE, ONEWIRE_OP_WRITE,
                                SEARCH_ROM_COMMAND, 8);
        //
        // Wait for completion of operation
        //
        while((OneWireIntStatus(ONEWIRE0_BASE, false)
            & ONEWIRE_INT_OP_DONE) != ONEWIRE_INT_OP_DONE);

        //
        // Clear ONEWIRE_INT_OP_DONE interrupt
        //
        OneWireIntClear(ONEWIRE0_BASE, ONEWIRE_INT_OP_DONE);

        //
        // Reset ui64SlaveAddr for storing new ROM number
        //
        ui64SlaveAddr = 0;

        //
        // In each iteration two bits are read and a bit is written to
        // obtain one bit of the ROM number.
        // This is done 64 times to obtain a complete ROM number.
        //
        for(ui32BitNumber = 0 ; ui32BitNumber < 64 ; ui32BitNumber++)
        {
            //
            // Reading logical AND of ith bits and logical AND of their
            // complements from the slaves.
            //
            OneWireTransaction(ONEWIRE0_BASE, ONEWIRE_OP_READ,
                                SEARCH_ROM_COMMAND, 2);

            //
            // Wait for completion of operation
            //
            while((OneWireIntStatus(ONEWIRE0_BASE, false)
                & ONEWIRE_INT_OP_DONE) != ONEWIRE_INT_OP_DONE);

            //
            // Clear ONEWIRE_INT_OP_DONE interrupt
            //
            OneWireIntClear(ONEWIRE0_BASE, ONEWIRE_INT_OP_DONE);

            //
            // Storing the read values in ui32ReadBits.
            //
            OneWireDataGet(ONEWIRE0_BASE, &ui32ReadBits);

            //
            // rejecting all other except least 2 bits
            //
            ui32ReadBits &= 3;

            //
            // Both bits zeroes ==> There is a conflict
            // and choice of path has to be made.
            //
            if(ui32ReadBits == 0)
            {
                if(ui32BitNumber == i32ConflictBitNumber)
                {
                    //
                    // Take path 1
                    // Writing 1 onto the bus
                    //
                    OneWireTransaction(ONEWIRE0_BASE, ONEWIRE_OP_WRITE,
                                        TAKE_PATH1, 1);

                    //
                    // Wait for completion of operation
                    //
                    while((OneWireIntStatus(ONEWIRE0_BASE, false)
                        & ONEWIRE_INT_OP_DONE) != ONEWIRE_INT_OP_DONE);

                    //
                    // Clear ONEWIRE_INT_OP_DONE interrupt
                    //
                    OneWireIntClear(ONEWIRE0_BASE , ONEWIRE_INT_OP_DONE);

                    //
                    // Storing the path taken
                    // (ui32BitNumber th bit of ROM number)
                    //
                    ui64SlaveAddr |= ((uint64_t)1 << ui32BitNumber);
                }
                else if(ui32BitNumber > i32ConflictBitNumber)
                {

                    //
                    // Take path 0
                    // Writing 0 onto the bus and wait for completion
                    //
                    OneWireTransaction(ONEWIRE0_BASE, ONEWIRE_OP_WRITE,
                         TAKE_PATH0 , 1 );

                    //
                    // Wait for completion of operation
                    //
                    while((OneWireIntStatus(ONEWIRE0_BASE, false) &
                        ONEWIRE_INT_OP_DONE) != ONEWIRE_INT_OP_DONE);

                    //
                    // Clear ONEWIRE_INT_OP_DONE interrupt
                    //
                    OneWireIntClear(ONEWIRE0_BASE, ONEWIRE_INT_OP_DONE);

                    //
                    // Updating i32LastConflictZeroBitNumber.
                    // Since there is a conflict and we are taking 0
                    //
                    i32LastConflictZeroBitNumber = ui32BitNumber;
                }
                else
                {
                    //
                    // Take the path of the PREVIOUS ROM number.
                    // Get the last path from last ROM number found
                    //
                    ui32WriteBit = ((g_pui64SlaveAddr[i32NumSlaves-1]
                                                 >> ui32BitNumber) & 1);

                    //
                    // Update ui64SlaveAddr with the path to be taken
                    //
                    ui64SlaveAddr |= ((uint64_t)ui32WriteBit <<
                            ui32BitNumber);

                    //
                    // Writing the previous path bit onto the bus
                    // and wait for completion
                    //
                    OneWireTransaction(ONEWIRE0_BASE, ONEWIRE_OP_WRITE,
                            ui32WriteBit, 1);

                    //
                    // Wait for completion of operation
                    //
                    while((OneWireIntStatus(ONEWIRE0_BASE, false)
                        & ONEWIRE_INT_OP_DONE) != ONEWIRE_INT_OP_DONE);

                    //
                    // Clear ONEWIRE_INT_OP_DONE interrupt
                    //
                    OneWireIntClear(ONEWIRE0_BASE, ONEWIRE_INT_OP_DONE);


                    if(ui32WriteBit == 0)
                    {
                        //
                        // Updating i32LastConflictZeroBitNumber.
                        // Since there is a conflict and we are taking 0
                        //
                        i32LastConflictZeroBitNumber = ui32BitNumber;
                    }
                }
            }
            else if(ui32ReadBits == 1)
            {
                //
                // Update ui64SlaveAddr with the path to be taken,
                // which is 1.
                //
                ui64SlaveAddr |= ((uint64_t)1 << ui32BitNumber);

                //
                // writing 1 onto the bus and wait for completion
                //
                OneWireTransaction(ONEWIRE0_BASE, ONEWIRE_OP_WRITE,
                                    TAKE_PATH1, 1);

                //
                // Wait for completion of operation
                //
                while((OneWireIntStatus(ONEWIRE0_BASE, false)
                     & ONEWIRE_INT_OP_DONE) != ONEWIRE_INT_OP_DONE);

                //
                // Clear ONEWIRE_INT_OP_DONE interrupt
                //
                OneWireIntClear(ONEWIRE0_BASE, ONEWIRE_INT_OP_DONE);

            }
            //
            // All slaves have this bit as 0
            //
            else
            {
                //
                // writing 0 onto the bus and wait for completion
                //
                OneWireTransaction(ONEWIRE0_BASE, ONEWIRE_OP_WRITE,
                        TAKE_PATH0 , 1);

                //
                // Wait for completion of operation
                //
                while((OneWireIntStatus(ONEWIRE0_BASE, false)
                     & ONEWIRE_INT_OP_DONE) != ONEWIRE_INT_OP_DONE);

                //
                // Clear ONEWIRE_INT_OP_DONE interrupt
                //
                OneWireIntClear(ONEWIRE0_BASE, ONEWIRE_INT_OP_DONE);
            }
        }

        //
        // Check CRC before storing the address in g_pui64SlaveAddr
        //
        if(!(OneWireCrcCheck(ui64SlaveAddr)))
        {
            //
            // return 0 => search fail
            //
            return 0;
        }

        //
        // Save the ROM number and increment number of slaves by 1
        //
        g_pui64SlaveAddr[i32NumSlaves] = ui64SlaveAddr;
        i32NumSlaves++;

        //
        // Update i32ConflictBitNumber after finding each ROM number.
        //
        i32ConflictBitNumber = i32LastConflictZeroBitNumber;

    }

    return i32NumSlaves ;
}

//*****************************************************************************
//
// Configure the UART before calling UARTprintf().
//
//*****************************************************************************
void
ConfigureUART(uint32_t ui32SysClock)
{
    //
    // Enable UART0
    //
    SysCtlPeripheralEnable(SYSCTL_PERIPH_UART0);

    //
    // Initialize the UART for console I/O.
    //
    UARTStdioConfig(0, 115200, ui32SysClock);
}

//*****************************************************************************
//
// This is the main application entry function.
//
//*****************************************************************************
int
main(void)
{
    uint32_t ui32SysClock;
    uint8_t  ui8Index;
    uint32_t ui32NumberofDevices = 0;
    static volatile uint32_t *pui32Address;

    //
    // Connfigure the System Clock
    //
    ui32SysClock = SysCtlClockFreqSet((SYSCTL_XTAL_25MHZ | SYSCTL_OSC_MAIN |
            SYSCTL_USE_PLL | SYSCTL_CFG_VCO_480), 120000000);

    //
    // Configure the pin out
    //
    PinoutSet();

    //
    // Enable Pad Pull
    //
    HWREG(GPIO_PORTG_BASE+GPIO_O_PUR) |= GPIO_PIN_4;

    //
    // Configure Serial Console
    //
    ConfigureUART(ui32SysClock);

    //
    // Disable the clock, reset the peripheral and then enable clock for
    // One-Wire module. Wait for the peripheral to be ready before performing
    // any access.
    //
    SysCtlPeripheralDisable(SYSCTL_PERIPH_ONEWIRE0);
    SysCtlPeripheralReset(SYSCTL_PERIPH_ONEWIRE0);
    SysCtlPeripheralEnable(SYSCTL_PERIPH_ONEWIRE0);

    while(!(SysCtlPeripheralReady(SYSCTL_PERIPH_ONEWIRE0)))
    {
    }

    //
    // Print the exmaple details
    //
    UARTprintf("\033[2JTM4C129x On Wire Enumeration Example\n");
    UARTprintf("------------------------------------\n\n");

    //
    // Disabiling all interrupts, not using any intterupts in this program
    //
    OneWireIntDisable(ONEWIRE0_BASE, ONEWIRE_IM_RST | ONEWIRE_IM_OPC |
                         ONEWIRE_IM_NOATR | ONEWIRE_IM_STUCK);

    //
    // Reset and wait for completion
    //
    OneWireBusReset(ONEWIRE0_BASE);

    //
    // Wait for the completion of reset
    //
    while((OneWireIntStatus(ONEWIRE0_BASE, false)
                     & ONEWIRE_INT_RESET_DONE) == 0);

    //
    // Clear ONEWIRE_INT_RESET_DONE interrupt
    //
    OneWireIntClear(ONEWIRE0_BASE, ONEWIRE_INT_RESET_DONE);

    //
    // Perform Enumeration
    //
    ui32NumberofDevices = OneWireEnum();

    //
    // reset and wait for completion
    //
    OneWireBusReset(ONEWIRE0_BASE);

    //
    // Wait for the completion of reset
    //
    while((OneWireIntStatus(ONEWIRE0_BASE, false)
                         & ONEWIRE_INT_RESET_DONE) == 0);

    //
    // Clear ONEWIRE_INT_RESET_DONE interrupt
    //
    OneWireIntClear(ONEWIRE0_BASE, ONEWIRE_INT_RESET_DONE);

    UARTprintf("Enumeration Completed... Found %d devices\n",
            ui32NumberofDevices);

    pui32Address = (uint32_t *)g_pui64SlaveAddr;

    for(ui8Index = 0; ui8Index < ui32NumberofDevices; ui8Index++)
    {
        UARTprintf("Device Code %d = %08x%08x\n", ui8Index,
                pui32Address[(ui8Index*2)+1], pui32Address[ui8Index*2]);

    }

    while(1);
}

